using JuMP, Ipopt
using CSV, DataFrames
using LinearAlgebra
using Plots

function load_data()
    node_data = CSV.read("data/feeder_15_node.csv", DataFrame)
    line_data = CSV.read("data/feeder_15_line.csv", DataFrame)
    der__data = CSV.read("data/feeder_15_der_.csv", DataFrame)

    # feeder size
    n = size(node_data,1)
    e = size(line_data,1)

    # line parameters
    r = line_data[:,:r]
    x = line_data[:,:x]
    node_f = line_data[:,:node_f]
    node_t = line_data[:,:node_t]

    # load data 
    d_p = node_data[2:end,:d_P]
    d_q = node_data[2:end,:d_Q] 

    # DER data 
    q̲ = der__data[:,:g_q_min] 
    q̅ = der__data[:,:g_q_max] 

    # Branch-bus-incidence-matrix
    Ã = zeros(e,n)
    for j in 1:e
        Ã[j,node_f[j]+1] = 1
        Ã[j,node_t[j]+1] = -1
    end
    a_0 = Ã[:,1]
    A = Ã[:,2:end]

    # LinDistFlow sensitivities
    R = 2*inv(A)*diagm(r)*inv(A)'
    X = 2*inv(A)*diagm(x)*inv(A)'

    return Dict(:R => R, :X => X, :d_p => d_p, :d_q => d_q, :q̲ => q̲, :q̅ => q̅,
                :v̅ => 1.05^2, :v̲ => 0.95^2, :n => n, :e => e, :A => A, :Ã => Ã)
end

data = load_data()

function DistOPF(data)
    model = Model(optimizer_with_attributes(Ipopt.Optimizer))
    JuMP.set_silent(model)
    # voltage
    @variable(model, v[1:data[:e]])
    # DER reactive power injection
    @variable(model, q[1:data[:e]])
    # objective function value 
    @objective(model, Min, 1/2*q'*q)
    # DER injection limits 
    @constraint(model, q .<= data[:q̅])
    @constraint(model, q .>= data[:q̲])
    # voltage limits 
    @constraint(model, v .>= data[:v̲])
    @constraint(model, v .<= data[:v̅])
    # voltage definition
    @constraint(model, v .== 1 .+ data[:R]*data[:d_p] .+ data[:X]*(data[:d_q].+q))
    # solve model
    optimize!(model)
    @info("Distribution OPF terminates with status $(termination_status(model))")
    return Dict(:v => JuMP.value.(v), :q => JuMP.value.(q), :obj => JuMP.objective_value(model))
end

sol = DistOPF(data)

function v_update(q,λ)
    model = Model(optimizer_with_attributes(Ipopt.Optimizer))
    JuMP.set_silent(model)
    # voltage
    @variable(model, v[1:data[:e]])
    # objective function value 
    @objective(model, Min, - λ'*v + ρ/2*(1 .+ data[:R]*data[:d_p] .+ data[:X]*(data[:d_q].+q) .- v)'*(1 .+ data[:R]*data[:d_p] .+ data[:X]*(data[:d_q].+q) .- v))
    # voltage limits 
    @constraint(model, v .>= data[:v̲])
    @constraint(model, v .<= data[:v̅])
    # solve model
    optimize!(model)
    return JuMP.value.(v)
end

function q_update(v,λ)
    model = Model(optimizer_with_attributes(Ipopt.Optimizer))
    JuMP.set_silent(model)
    # DER reactive power injection
    @variable(model, q[1:data[:e]])
    # objective function value 
    @objective(model, Min, 1/2*q'*q + λ'*data[:X]'q + ρ/2*(1 .+ data[:R]*data[:d_p] .+ data[:X]*(data[:d_q].+q) .- v)'*(1 .+ data[:R]*data[:d_p] .+ data[:X]*(data[:d_q].+q) .- v))
    # DER injection limits 
    @constraint(model, q .<= data[:q̅])
    @constraint(model, q .>= data[:q̲])
    # solve model
    optimize!(model)
    return JuMP.value.(q)
end

function λ_update(v,q,λ)
    return λ + ρ * (1 .+ data[:R]*data[:d_p] .+ data[:X]*(data[:d_q].+q) .- v)
end

K = 10000
ρ = 1
ε_tol = 10e-10

λ = zeros(data[:e])
v = ones(data[:e])
q = zeros(data[:e])
res = zeros(K)

@time for k in 1:K
    v = v_update(q,λ)
    q = q_update(v,λ)
    λ = λ_update(v,q,λ)
    res[k] = norm(1 .+ data[:R]*data[:d_p] .+ data[:X]*(data[:d_q].+q) .- v)
    if res[k] <= ε_tol
        println("ADMM converged in $k iterations") 
        break 
    end
    if k == K && res[k] > ε_tol
        println("ADMM did not converge in $K iterations")
    end
end

# Compare the solutions (unsquared voltage)
sol_comp = DataFrame(v_cent = Any[], v_admm = Any[], Δv = Any[], q_cent = Any[], q_admm = Any[], Δq = Any[])
for i in 1:data[:e]
    push!(sol_comp,
            [round(sqrt.(sol[:v])[i],digits=3),
             round(sqrt.(v)[i],digits=3),
             round(norm(sqrt.(sol[:v])[i].-sqrt.(v)[i]),digits=3),
             round(sol[:q][i],digits=3),
             round(q[i],digits=3),
             round(norm(sol[:q][i].-q[i]),digits=3)])
end
display(sol_comp)

plot(res[findall(x->x!=0,res)],yaxis=:log,frame=:box,xlabel="iteration k",ylabel="primal residual (log scale)",label=false)